#ifndef UTIL_H
#define UTIL_H

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <json-c/json.h>

#include "client.h"

#define WM_VERT_LEFT  0
#define WM_VERT_RIGHT 1
#define WM_HORIZ_TOP  0
#define WM_HORIZ_BOT  1

typedef enum NodeType {
    NODE_VERTICAL,
    NODE_HORIZONTAL,
    NODE_TAB,
    NODE_CLIENT,
    NODE_COUNT
} NodeType;

typedef struct TreeNode TreeNode;

typedef struct {
    TreeNode **nodes;
    size_t size;
    size_t capacity;
} NodeArray;

typedef struct {
    uint64_t *elements;
    size_t size;
    size_t capacity;
} UIntArray;

typedef struct {
    int x;
    int y;
    int w;
    int h;
} Rect;

struct TreeNode {
    NodeType type;

    TreeNode *parent;

    NodeArray *children;

    Rect pos;

    Client *client;
    unsigned int id;
    uint64_t distance;
};

typedef struct {
    size_t ws_index;
    char *str;
} WmWorkspaceToStrRet;

typedef struct {
    time_t timestamp;
    bool after;
    char function_name[128];
    NodeArray *workspaces;
} LogEntry;

typedef uint64_t (*UpdateCostFunction)(TreeNode*, TreeNode*);

typedef struct {
    uint64_t insert_cost;
    uint64_t remove_cost;
    UpdateCostFunction update_cost_function;
} TreeEditDistanceCosts;

/**
 * Functions for NodeArray operations
 */

/**
 * Initializes a new NodeArray.
 */
NodeArray* wm_nodearray_new();

/**
 * Adds `node` to the end of `arr`. If the capacity is lower than the
 * new size, it resizes the array.
 */
void wm_nodearray_push(NodeArray *arr, TreeNode *node);

/**
 * Removes the last element, and returns it in `ret`.
 */
bool wm_nodearray_pop(NodeArray *arr, TreeNode **ret);

/**
 * Removes the first element, and returns it in `ret`.
 */
bool wm_nodearray_pop_front(NodeArray *arr, TreeNode **ret);

/**
 * Removes all elements from `arr`, and sets its size to 0.
 */
void wm_nodearray_clear(NodeArray *arr);

/**
 * Removes the element at the provided index. Returns true on success, false
 * on failure.
 */
bool wm_nodearray_remove(NodeArray *arr, size_t index);

/**
 * Frees the array.
 */
void wm_nodearray_free(NodeArray *arr);

/**
 * Functions for TreeNode operations
 */

/**
 * Initializes a new TreeNode.
 */
TreeNode* wm_treenode_new(NodeType type, TreeNode *parent);

/**
 * Frees `node`, and the `children` NodeArray field.
 */
void wm_treenode_free(TreeNode *node);

/**
 * Returns true if `node` has no children.
 */
bool wm_treenode_is_empty(TreeNode *node);

/**
 * Splits the available space between `ret1` and `ret2`, based on `node->type`.
 */
void wm_treenode_split_space(TreeNode *node, Rect *ret1, Rect *ret2);

/**
 * Asserts `node->children->size == 2`, and calls split_space on first and second
 * child.
 */
void wm_treenode_recalculate_space(TreeNode *node);

/**
 * Swaps the clients of the `node1` and `node2`.
 */
void wm_treenode_swap(TreeNode *root, TreeNode *node1, TreeNode* node2);

/**
 * Pushes child to `node->children`, and sets `child->parent` to node.
 */
void wm_treenode_add_child(TreeNode *node, TreeNode *child);

/**
 * Finds `client` in `root`, removes and frees the node associated with it,
 * and recalculates layout.
 */
TreeNode* wm_treenode_remove_client(Wm *wm, TreeNode *root, Client *client);

/**
 * Finds `node` in `root`, removes and frees it, and recalculates layout.
 */
void wm_treenode_remove_node(Wm *wm, TreeNode *root, TreeNode *node);

/**
 * Converts `tree` to DOT file format.
 * Use `dot <filename> -Tsvg > output.svg` to convert the DOT file to svg.
 */
void wm_tree_to_DOT(TreeNode *root, const char *filename);

/**
 * Converts `type` to a string, and writes it to `buf`.
 */
void wm_node_type_to_str(NodeType type, char *buf, size_t bufsize);

/**
 * Finds non empty workspaces, and converts them to JSON strings.
 * Return value: `UIntArray<WmWorkspaceToStrRet*>`
 */
UIntArray* wm_nonempty_workspaces_to_strptrarray(Wm *wm);

/**
 * Converts a string to NodeType.
 */
NodeType wm_node_type_from_str(char *buf, size_t bufsize);

/**
 * Converts `node` to a JSON string.
 */
char* wm_treenode_to_str(TreeNode *node);

/**
 * Converts `node` to a JSON string, and if DEBUG is defined, prints it to stderr.
 */
void wm_treenode_print(TreeNode *node);

/**
 * Parses a `json_object` to a TreeNode.
 */
TreeNode* wm_json_obj_to_treenode(struct json_object *jobj);


/**
 * Functions for UIntArray operations
 * UIntArray can be used as a generic pointer container, if
 * `_Static_assert(sizeof(uint64_t) == sizeof(<pointer_type>), "")`.
 */

/**
 * Initializes a new UIntArray.
 */
UIntArray* wm_uintarray_new();

/**
 * Asserts `index < arr->size`, and returns `arr->elements[index]`.
 */
uint64_t wm_uintarray_at(UIntArray *arr, size_t index);

/**
 * Adds `node` to the end of `arr`. If the capacity is lower than the
 * new size, it resizes the array.
 */
void wm_uintarray_push(UIntArray *arr, uint64_t element);

/**
 * Removes the last element, and returns it in `ret`.
 */
bool wm_uintarray_pop(UIntArray *arr, uint64_t *ret);

/**
 * Removes the first element, and returns it in `ret`.
 */
bool wm_uintarray_pop_front(UIntArray *arr, uint64_t *ret);

/**
 * Removes the element at the provided index. Returns true on success, false
 * on failure.
 */
bool wm_uintarray_remove(UIntArray *arr, size_t index);

/**
 * Frees the array.
 */
void wm_uintarray_free(UIntArray *arr);

/**
 * Returns ((uint64_t**)arr->elements)[i][j].
 */
uint64_t wm_uintarray_2d_index(UIntArray *arr, size_t i, size_t j);

/**
 * TreeNode utility functions
 */

/**
 * Find the node associated with the currently focused client, and returns it.
 * Returns NULL if no client is focused.
 */
TreeNode* wm_treenode_ptr_find_focused_client_node(TreeNode *root);

/**
 * Find the node associated with `client`, and returns it.
 * Returns NULL if it could not find `client`.
 */
TreeNode* wm_treenode_ptr_find_client_node(TreeNode *root, Client *client);

/**
 * Returns all `NODE_CLIENT` type descendant nodes of `root`.
 */
NodeArray* wm_treenode_find_client_nodes(TreeNode *root);

/**
 * Returns the sibling node of `node`.
 * Returns NULL on error.
 */
TreeNode* wm_treenode_split_get_sibling(TreeNode *node);

/**
 * Returns descendant nodes of `root`.
 */
NodeArray* wm_all_nodes_to_array(TreeNode *root);

/**
 * Converts all non empty workspace trees to a `json_object`.
 */
json_object* wm_state_to_json_object(Wm *wm);

/**
 * Writes current state encoded as JSON to logfile.
 * If logfile exists, and is not empty, appends the state to the JSON object in
 * logfile.
 */
void wm_log_state(Wm *wm, const char *prefixstr, const char* logfile);


/**
 * Utility functions for tree edit distance
 */

/**
 * Returns the leftmost descendant of `node`.
 */
TreeNode* wm_treenode_lmd(TreeNode *node);

/**
 * Returns all leftmost descendants of `node`.
 */
NodeArray* wm_treenode_all_lmds(TreeNode *node);

/**
 * Returns all leftmost descendants of `node`, as indexes of a postorder
 * traversal array.
 */
UIntArray* wm_treenode_all_lmds_index(TreeNode *node);

/**
 * Returns all nodes of `tree`, in postorder traversal order.
 */
NodeArray* wm_postorder_traversal(TreeNode *tree);

/**
 * Returns true of `node` is a keyroot.
 * LR_keyroots(T) = {k | there exists no k’ > k such that l(k)= l(k’)}.
 * l = leftmost descendant
 */
bool wm_is_treenode_keyroot(TreeNode *node, NodeArray *postorder);

/**
 * Returns all keyroots of `node`.
 */
NodeArray* wm_treenode_all_keyroots(TreeNode *node);

/**
 * Returns all keyroots of `node`, as indexes of a postorder traversal array.
 */
UIntArray* wm_treenode_all_keyroots_index(TreeNode *node);


/**
 * Calculates the distance of the subtrees of `tree1` and `tree2`, at indexes
 * `i` and `j`.
 */
void wm_treedist(TreeNode *tree1, TreeNode *tree2, size_t i, size_t j,
                 UIntArray *dists, TreeEditDistanceCosts costs);

/**
 * Calculates the distance of `tree1` and `tree2` using the edit costs `costs`.
 */
size_t wm_tree_edit_distance(TreeNode *tree1, TreeNode *tree2, TreeEditDistanceCosts costs);

/**
 * Creates the logfile at the provided path.
 */
void wm_logfile_init(const char *path);

/**
 * Frees all dynamically allocated fields of the array elements, and frees the
 * array.
 * Type of `entries`: UIntArray<LogEntry*>
 */
void wm_logentries_free(UIntArray *entries);

/**
 * Parses the JSON objects in the file at path `filename`, and populates a LogEntry*
 * array with the parsed values.
 * Return type: UIntArray<LogEntry*>
 */
UIntArray* wm_read_log(const char* filename);

/**
 * Calculates the tree edit distance of `log_entries[i]` and `curr_tree`.
 * Stores the distance in `log_entries[i]->workspaces[j]->tree->distance`.
 * Type of `log_entries`: UIntArray<LogEntry*>
 */
void wm_logentries_calculate_distances(UIntArray *log_entries, TreeNode *curr_tree,
                                       UpdateCostFunction update_cost_function);

#endif
